-------------------------------------------------------------------------------
-- hicolvga.vhd
--
-- Author(s):     Jorgen Peddersen and Ashley Partis
-- Created:       Jan 2001
-- Last Modified: Jan 2001
-- 
-- This code acts as a top level for the VGA output project.  The RAMDAC should
-- be set up to program high colour dual-edged mode.  This displays a pattern
-- of horizontal and vertical bands of colour which can have their colour
-- changed and rotated around the screen.  Experimentation is the best way to
-- understand what each switch does.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity vga is
    port (
        clk: in STD_LOGIC;								-- clock
        rstn: in STD_LOGIC;								-- asynchronous active low reset
        red: in STD_LOGIC_VECTOR(1 downto 0);			-- switches to control the red component of the test pattern
        green: in STD_LOGIC_VECTOR(1 downto 0);			-- switches to control the green component of the test pattern
        blue: in STD_LOGIC_VECTOR(1 downto 0);			-- switches to control the blue component of the test pattern
        vertRotate: in STD_LOGIC;						-- if high, the pattern will rotate vertically
        horiRotate: in STD_LOGIC;						-- if high, the pattern will rotate horizontally
        vertDirection: in STD_LOGIC;					-- direction of vertical rotation
        horiDirection: in STD_LOGIC;					-- direction of horizontal rotation
        pixel: out STD_LOGIC_VECTOR (7 downto 0);		-- RAMDAC pixel lines
        blankn: out STD_LOGIC;							-- RAMDAC blank signal
        RDn: out STD_LOGIC;								-- RAMDAC RDn connection
		WRn: out STD_LOGIC;        						-- RAMDAC WRn connection
		RAMDACD: inout STD_LOGIC_VECTOR (7 downto 0);	-- RAMDAC data lines
		RS: inout STD_LOGIC_VECTOR (2 downto 0);		-- RAMDAC RS lines
        hsync: out STD_LOGIC;							-- horizontal sync for monitor
        vsync: out STD_LOGIC;							-- vertical sync for monitor
        triste: out STD_LOGIC;							-- signal to tristate ethernet PHY
        rramce: out STD_LOGIC;							-- right ram chip enable
        pixelclk: out STD_LOGIC							-- RAMDAC pixel clock
    );
end vga;

architecture vga_arch of vga is

-- control VGA signals
component vgacore 
	generic (
		H_SIZE : integer;							-- horizontal size of input image, MAX 800
		V_SIZE : integer							-- vertical size of input image, MAX 600
	);
	port
	(
		reset: in std_logic;						-- asynchronous active low reset
		clock: in std_logic;						-- clock
		hsyncb: buffer std_logic;					-- horizontal (line) sync
		vsyncb: out std_logic;						-- vertical (frame) sync
		latch: out STD_LOGIC;						-- latches new rgb value
		enable: out STD_LOGIC;						-- enable/ground RGB output lines
		hloc: out std_logic_vector(9 downto 0);		-- horizontal address to be decoded for video RAM
		vloc: out std_logic_vector(9 downto 0)		-- vertical address to be decoded for video RAM
	);
end component;

-- Program the RAMDAC
component prgramdacver2
    port (
        clk: in STD_LOGIC;							-- Clock
        rstn: in STD_LOGIC;							-- Asynchronous active low reset
        start: in STD_LOGIC;						-- Start signal
        done: out STD_LOGIC;						-- Asserted when programming is finished
        WRn: out STD_LOGIC;							-- Write line to RAMDAC (active low)
        RDn: out STD_LOGIC;							-- Read line to RAMDAC (active low)
        RS: inout STD_LOGIC_VECTOR (2 downto 0);	-- Register select lines to the RAMDAC
        data: inout STD_LOGIC_VECTOR (7 downto 0)	-- Bidirectional data line to RAMDAC
    );
end component;

--game of life module
component life
    port (
        clk: in STD_LOGIC;								-- clock
		outArray : out std_logic_vector(8191 downto 0);	--output buffer
		iteration : out INTEGER 						--loop count
    );
end component;


-- State signals
type STATETYPE is (stReset, stWaste1, stWaste2, stWaste3, stWait, stForever);
signal presState: STATETYPE;


-- signals so that hsync and vsync can be read
signal hsyncInt : STD_LOGIC;
signal vsyncInt : STD_LOGIC;

-- signals to cue different processes
signal startProg : STD_LOGIC;
signal startVGA : STD_LOGIC;
signal resetVGA : STD_LOGIC;
signal done : STD_LOGIC;

-- signals to generate test pattern
signal hloc : std_logic_vector(9 downto 0);				-- horizontal location of each pixelach pixel
signal vloc : std_logic_vector(9 downto 0);				-- vertical location of each pixel
signal vertoffset : std_logic_vector(8 downto 0);		-- offset for vertical rotation
signal horioffset : std_logic_vector(8 downto 0);		-- offset for horizontal rotation
signal vertVal : std_logic_vector(8 downto 0);			-- adds offset to pixel location
signal horiVal : std_logic_vector(8 downto 0);			-- adds offset to pixel location
signal pixelData : STD_LOGIC_VECTOR(15 downto 0);		-- colour to write for each pixel

signal iteration : integer := 0;						--loop count
signal outArray : std_logic_vector(8191 downto 0) := (others => '0');
type TWO_D_ARRAY is array(63 downto 0) of std_logic_vector(127 downto 0);
signal framebuffer : TWO_D_ARRAY;


begin
	-- VGA controller
	cycler : vgacore 
	generic map (
		H_SIZE => 512,
		V_SIZE => 512
	)
	port map(
		reset => resetVGA,
		clock => clk,
		hsyncb => hsyncInt,
		vsyncb => vsyncInt,
		latch => open,
		enable => blankn,
		hloc => hloc,
		vloc => vloc
	);
    
    -- RAMDAC programmer
    RAMDACprog : prgramdacver2 port map (
        clk => clk,
        rstn => rstn,
        start => startProg,
        done => done,
        WRn => WRn,
        RDn => RDn,
        RS => RS,
        data => RAMDACD
    );

	--life controller
	lifegame : life port map (
		clk => clk,
		outArray => outArray,
		iteration => iteration
	);
	
	-- This is a simple mealy state machine that cues the VGA controller
	-- when the RAMDAC is finished programming
	process(clk, rstn)
	begin
		if rstn = '0' then
			presState <= stReset;
		elsif clk'event AND clk = '1' then
			case presState is
				when stReset =>
					presState <= stWaste1;
				when stWaste1 =>
					presState <= stWaste2;
				when stWaste2 =>
					presState <= stWaste3;
				when stWaste3 =>
					presState <= stWait;
				when stWait =>
					if done = '0' then
						presState <= stWait;
					else
						presState <= stForever;
					end if;
				when stForever =>
					presState <= stForever;
			end case;
		end if;
	end process;
	
	process(presState)
	begin
		case presState is
			when stReset =>
				startProg <= '1';
				startVGA <= '0';
			when stWaste1 =>
				startProg <= '1';
				startVGA <= '0';
			when stWaste2 =>
				startProg <= '1';
				startVGA <= '0';
			when stWaste3 =>
				startProg <= '1';
				startVGA <= '0';
			when stWait =>
				startProg <= '0';
				startVGA <= '0';
			when stForever =>
				startProg <= '0';
				startVGA <= '1';
		end case;
	end process;
	
	-- calculate offsets for data that is rotating.  Value will update whenever
	-- a new page is called for as it is clocked by vsync.
	process(rstn, vsyncInt)
	begin
		if rstn = '0' then
			vertoffset <= "000000000";
			horioffset <= "000000000";
		elsif vsyncInt'event and vsyncInt = '1' then
			-- update vertical counter if required
			if vertRotate = '1' then
				if vertDirection = '0' then
					vertoffset <= vertoffset + 1;
				else
					vertoffset <= vertoffset - 1;				
				end if;
			end if;
			-- update horizontal counter if required
			if horiRotate = '1' then
				if horiDirection = '0' then
					horioffset <= horioffset + 1;
				else
					horioffset <= horioffset - 1;				
				end if;
			end if;
		end if;
	end process;
	
	pixelData(15) <= '0';

	process(clk)
	begin
		if clk'event and clk = '1' then
			--64 128-bit copies to outArray
			framebuffer(0) <= outArray(127 downto 0);
			framebuffer(1) <= outArray(255 downto 128);
			framebuffer(2) <= outArray(383 downto 256);
			framebuffer(3) <= outArray(511 downto 384);
			framebuffer(4) <= outArray(639 downto 512);
			framebuffer(5) <= outArray(767 downto 640);
			framebuffer(6) <= outArray(895 downto 768);
			framebuffer(7) <= outArray(1023 downto 896);
			framebuffer(8) <= outArray(1151 downto 1024);
			framebuffer(9) <= outArray(1279 downto 1152);
			framebuffer(10) <= outArray(1407 downto 1280);
			framebuffer(11) <= outArray(1535 downto 1408);
			framebuffer(12) <= outArray(1663 downto 1536);
			framebuffer(13) <= outArray(1791 downto 1664);
			framebuffer(14) <= outArray(1919 downto 1792);
			framebuffer(15) <= outArray(2047 downto 1920);
			framebuffer(16) <= outArray(2175 downto 2048);
			framebuffer(17) <= outArray(2303 downto 2176);
			framebuffer(18) <= outArray(2431 downto 2304);
			framebuffer(19) <= outArray(2559 downto 2432);
			framebuffer(20) <= outArray(2687 downto 2560);
			framebuffer(21) <= outArray(2815 downto 2688);
			framebuffer(22) <= outArray(2943 downto 2816);
			framebuffer(23) <= outArray(3071 downto 2944);
			framebuffer(24) <= outArray(3199 downto 3072);
			framebuffer(25) <= outArray(3327 downto 3200);
			framebuffer(26) <= outArray(3455 downto 3328);
			framebuffer(27) <= outArray(3583 downto 3456);
			framebuffer(28) <= outArray(3711 downto 3584);
			framebuffer(29) <= outArray(3839 downto 3712);
			framebuffer(30) <= outArray(3967 downto 3840);
			framebuffer(31) <= outArray(4095 downto 3968);
			framebuffer(32) <= outArray(4223 downto 4096);
			framebuffer(33) <= outArray(4351 downto 4224);
			framebuffer(34) <= outArray(4479 downto 4352);
			framebuffer(35) <= outArray(4607 downto 4480);
			framebuffer(36) <= outArray(4735 downto 4608);
			framebuffer(37) <= outArray(4863 downto 4736);
			framebuffer(38) <= outArray(4991 downto 4864);
			framebuffer(39) <= outArray(5119 downto 4992);
			framebuffer(40) <= outArray(5247 downto 5120);
			framebuffer(41) <= outArray(5375 downto 5248);
			framebuffer(42) <= outArray(5503 downto 5376);
			framebuffer(43) <= outArray(5631 downto 5504);
			framebuffer(44) <= outArray(5759 downto 5632);
			framebuffer(45) <= outArray(5887 downto 5760);
			framebuffer(46) <= outArray(6015 downto 5888);
			framebuffer(47) <= outArray(6143 downto 6016);
			framebuffer(48) <= outArray(6271 downto 6144);
			framebuffer(49) <= outArray(6399 downto 6272);
			framebuffer(50) <= outArray(6527 downto 6400);
			framebuffer(51) <= outArray(6655 downto 6528);
			framebuffer(52) <= outArray(6783 downto 6656);
			framebuffer(53) <= outArray(6911 downto 6784);
			framebuffer(54) <= outArray(7039 downto 6912);
			framebuffer(55) <= outArray(7167 downto 7040);
			framebuffer(56) <= outArray(7295 downto 7168);
			framebuffer(57) <= outArray(7423 downto 7296);
			framebuffer(58) <= outArray(7551 downto 7424);
			framebuffer(59) <= outArray(7679 downto 7552);
			framebuffer(60) <= outArray(7807 downto 7680);
			framebuffer(61) <= outArray(7935 downto 7808);
			framebuffer(62) <= outArray(8063 downto 7936);
			framebuffer(63) <= outArray(8191 downto 8064);
		end if;
	end process;

		-- draw current array
	process(clk)
		variable v_round : INTEGER;
		variable h_round : INTEGER;
	begin
		if clk'event and clk = '1' then
			v_round := conv_integer(vloc(8 downto 3));
			h_round := conv_integer(hloc(8 downto 2));
			if framebuffer(v_round)(h_round) = '1' then
				pixelData(14 downto 0) <= "111111111111111";
			else
				pixelData(14 downto 0) <= "000000000000000";
			end if;
		end if;
	end process;

	-- handle dual-edged clock to give correct data to the RAMDAC
	pixel <= pixelData(7 downto 0) when clk = '1' else pixelData(15 downto 8);
	
	-- only start VGA after RAMDAC has been programmed
	resetVGA <= rstn AND startVGA;
	
	-- pass the outputs out
	hsync <= hsyncINT;
	vsync <= vsyncINT;
	
	-- Provide 50MHz pixel clock
	pixelclk <= clk;
	
	-- turn off the ethernet outputs and the right SRAM bank to avoid contention on the lines
	triste <= '1';
	rramce <= '1';

end vga_arch;
