--game of life module

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity life is
    port (
        clk: in STD_LOGIC;								-- clock
		outArray : out std_logic_vector(8191 downto 0);		--output buffer
		iteration : out INTEGER 						--loop count
    );
end life;

architecture life_arch of life is

--states in the game of life
type lifestate is (initializeSt, offsetsSt, getcellsSt, countcellsSt, applyrulesSt, copySt, delaySt, writecellSt);
signal gamestate : lifestate := initializeSt;

--state bits for our state machine
--variables that control which processes (datapath mechanisms) fire
signal initialize : BIT := '0';
signal getcells : BIT := '0';
signal offsets : BIT := '0';
signal applyrules : BIT := '0';
signal countcells : BIT := '0';
signal delay : BIT := '0';
signal writecell : BIT := '0';
signal copy_over : BIT := '0';

signal done_initialize, done_copy, done_offsets, done_getcells, done_countcells, done_applyrules, done_writecell, done_delay : bit := '0';

signal activeCells : INTEGER := 0;  --adjacent active cells

signal countDown : INTEGER := 0;
type TWO_D_ARRAY is array(63 downto 0) of std_logic_vector(127 downto 0);
type FULL_ARRAY is array(1 downto 0) of TWO_D_ARRAY;
signal gameArray : FULL_ARRAY;

subtype arrayIndex is integer range 0 to 1;
signal curr_array : arrayIndex := 0;
signal other_array : arrayIndex := 1;

signal loops : integer := 0;

--interface variables for rules module
signal cell0, cell1, cell2, cell3, cell4, cell5, cell6, cell7 : std_logic;
subtype index is integer range 0 to 126;
signal x : index := 0;
signal y : index := 1;
signal processedcell : std_logic;
signal currentcell : std_logic;

begin

	---@@@@@@@@@@@@@@@@@@@@@@@@@@@@ STATE MACHINE @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	--state control mechanism
	process(clk)
	begin 
		case gamestate is
			when initializeSt =>
				initialize <= '1'; --start initialization of the 2 bit arrays
				getcells <= '0'; --deny rules process
				offsets <= '0'; --deny offsets
				applyrules <= '0'; --deny rules
				countcells <= '0';  --deny counting of adjacent cells
				writecell <= '0';
				delay <= '0';	--deny delay
				copy_over <= '0';
			when offsetsSt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '1';  --calculate the x-y offsets
				applyrules <= '0';
				countcells <= '0';
				writecell <= '0';
				delay <= '0';
				copy_over <= '0';
			when getcellsSt =>
				initialize <= '0'; 
				getcells <= '1'; --start applying rules to the state arrays				
				offsets <= '0';
				applyrules <= '0';
				countcells <= '0';
				writecell <= '0';
				delay <= '0';
				copy_over <= '0';
			when countcellsSt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '0';
				applyrules <= '0';
				countcells <= '1';  --activate counting of adjacent cells
				writecell <= '0';
				delay <= '0';
				copy_over <= '0';
			when applyrulesSt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '0';
				applyrules <= '1';  --activate rules process
				countcells <= '0';
				writecell <= '0';
				delay <= '0';
				copy_over <= '0';
			when writecellSt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '0';
				applyrules <= '0';
				countcells <= '0';
				writecell <= '1';
				delay <= '0';	--activate delay process		
				copy_over <= '0';
			when copySt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '0';
				applyrules <= '0';
				countcells <= '0';
				writecell <= '0';
				delay <= '0';	--activate delay process		
				copy_over <= '1';
			when delaySt =>
				initialize <= '0';
				getcells <= '0';
				offsets <= '0';
				applyrules <= '0';
				countcells <= '0';
				writecell <= '0';
				delay <= '1';	--activate delay process
				copy_over <= '0';
		end case;
	end process;

	--this process controls the state transition depending upon what
	--process in the datapath has finished
	process(clk)
	begin 
		if clk'event and clk = '1' then
			if done_initialize = '1' then
				gamestate <= offsetsSt;
			elsif done_offsets = '1' then
				gamestate <= getcellsSt;
			elsif done_getcells = '1' then
				gamestate <= countcellsSt;
			elsif done_countcells = '1' then
				gamestate <= applyrulesSt;
			elsif done_applyrules = '1' then
				gamestate <= writecellSt;		
			elsif done_writecell = '1' then
				if x = 62 and y = 126 then
					gamestate <= copySt;
				else
					gamestate <= offsetsSt;
				end if;
			elsif done_copy = '1' then
				gamestate <= delaySt;
			elsif done_delay = '1' then
				gamestate <= offsetsSt;
			end if;
		end if;
	end process;

	process(delay, clk)
		variable rate_counter : INTEGER := 0;
	begin
		if clk'EVENT and clk = '1' then
	 		if delay = '1' then		-- set this when initializing gameArray finishes
				if rate_counter < 5 then
					rate_counter := rate_counter + 1;
				else
					done_delay <= '1';
					rate_counter := 0;
				end if;
			else
				done_delay <= '0';
			end if;
		end if;
	end process;

	process(offsets, clk)
	begin
		if clk'event and clk = '1' then
 			if offsets = '1' and done_offsets = '0' then
				if x < 62 then
					x <= x + 1;
				elsif y < 126 then
					y <= y + 1;
					x <= 1;
				elsif y = 126 then
					y <= 1;
					x <= 1;
					loops <= loops + 1;
					iteration <= loops;

					--64 128-bit copies to outArray
					outArray(127 downto 0) <= gameArray(0)(0);
					outArray(255 downto 128) <= gameArray(0)(1);
					outArray(383 downto 256) <= gameArray(0)(2);
					outArray(511 downto 384) <= gameArray(0)(3);
					outArray(639 downto 512) <= gameArray(0)(4);
					outArray(767 downto 640) <= gameArray(0)(5);
					outArray(895 downto 768) <= gameArray(0)(6);
					outArray(1023 downto 896) <= gameArray(0)(7);
					outArray(1151 downto 1024) <= gameArray(0)(8);
					outArray(1279 downto 1152) <= gameArray(0)(9);
					outArray(1407 downto 1280) <= gameArray(0)(10);
					outArray(1535 downto 1408) <= gameArray(0)(11);
					outArray(1663 downto 1536) <= gameArray(0)(12);
					outArray(1791 downto 1664) <= gameArray(0)(13);
					outArray(1919 downto 1792) <= gameArray(0)(14);
					outArray(2047 downto 1920) <= gameArray(0)(15);
					outArray(2175 downto 2048) <= gameArray(0)(16);
					outArray(2303 downto 2176) <= gameArray(0)(17);
					outArray(2431 downto 2304) <= gameArray(0)(18);
					outArray(2559 downto 2432) <= gameArray(0)(19);
					outArray(2687 downto 2560) <= gameArray(0)(20);
					outArray(2815 downto 2688) <= gameArray(0)(21);
					outArray(2943 downto 2816) <= gameArray(0)(22);
					outArray(3071 downto 2944) <= gameArray(0)(23);
					outArray(3199 downto 3072) <= gameArray(0)(24);
					outArray(3327 downto 3200) <= gameArray(0)(25);
					outArray(3455 downto 3328) <= gameArray(0)(26);
					outArray(3583 downto 3456) <= gameArray(0)(27);
					outArray(3711 downto 3584) <= gameArray(0)(28);
					outArray(3839 downto 3712) <= gameArray(0)(29);
					outArray(3967 downto 3840) <= gameArray(0)(30);
					outArray(4095 downto 3968) <= gameArray(0)(31);
					outArray(4223 downto 4096) <= gameArray(0)(32);
		  			outArray(4351 downto 4224) <= gameArray(0)(33);
					outArray(4479 downto 4352) <= gameArray(0)(34);
					outArray(4607 downto 4480) <= gameArray(0)(35);
					outArray(4735 downto 4608) <= gameArray(0)(36);
					outArray(4863 downto 4736) <= gameArray(0)(37);
					outArray(4991 downto 4864) <= gameArray(0)(38);
					outArray(5119 downto 4992) <= gameArray(0)(39);
					outArray(5247 downto 5120) <= gameArray(0)(40);
					outArray(5375 downto 5248) <= gameArray(0)(41);
					outArray(5503 downto 5376) <= gameArray(0)(42);
					outArray(5631 downto 5504) <= gameArray(0)(43);
					outArray(5759 downto 5632) <= gameArray(0)(44);
					outArray(5887 downto 5760) <= gameArray(0)(45);
					outArray(6015 downto 5888) <= gameArray(0)(46);
					outArray(6143 downto 6016) <= gameArray(0)(47);
					outArray(6271 downto 6144) <= gameArray(0)(48);
					outArray(6399 downto 6272) <= gameArray(0)(49);
					outArray(6527 downto 6400) <= gameArray(0)(50);
					outArray(6655 downto 6528) <= gameArray(0)(51);
					outArray(6783 downto 6656) <= gameArray(0)(52);
					outArray(6911 downto 6784) <= gameArray(0)(53);
					outArray(7039 downto 6912) <= gameArray(0)(54);
					outArray(7167 downto 7040) <= gameArray(0)(55);
					outArray(7295 downto 7168) <= gameArray(0)(56);
					outArray(7423 downto 7296) <= gameArray(0)(57);
					outArray(7551 downto 7424) <= gameArray(0)(58);
					outArray(7679 downto 7552) <= gameArray(0)(59);
					outArray(7807 downto 7680) <= gameArray(0)(60);
					outArray(7935 downto 7808) <= gameArray(0)(61);
					outArray(8063 downto 7936) <= gameArray(0)(62);
					outArray(8191 downto 8064) <= gameArray(0)(63);
				end if;			
				done_offsets <= '1';
			else
				done_offsets <= '0';
			end if;
		end if;
	end process;

--PREFETCH ADJACENT CELLS-----------------------------------------------------------------------

	--this process stores all adjacent cells in the in the following registers
	process(getcells, clk)
	begin
		if clk'event and clk = '1' then 
			if getcells = '1' then
				cell0 <= gameArray(curr_array)(x)(y-1);
				cell1 <= gameArray(curr_array)(x+1)(y-1);
				cell2 <= gameArray(curr_array)(x+1)(y);
				cell3 <= gameArray(curr_array)(x+1)(y+1);
				cell4 <= gameArray(curr_array)(x)(y+1);
				cell5 <= gameArray(curr_array)(x-1)(y+1);
				cell6 <= gameArray(curr_array)(x-1)(y);
				cell7 <= gameArray(curr_array)(x-1)(y-1);
				currentcell <= gameArray(curr_array)(x)(y);
				done_getcells <= '1';
	   	else
				done_getcells <= '0';
			end if;
		end if;
	end process;

	--this process counts adjacent cells of the current node
	process(countcells, clk)
		variable temp1, temp2 : INTEGER;
	begin
		if clk'event and clk = '1' then
			if countcells = '1' and done_countcells = '0' then
				temp1 := 0;	--reset cell count
				temp2 := 0;
				if cell0 = '1' then
					temp1 := temp1 + 1;
				end if;
				if cell1 = '1' then
					temp1 := temp1 + 1;
				end if;
				if cell2 = '1' then
					temp1 := temp1 + 1;
				end if;
				if cell3 = '1' then
					temp1 := temp1 + 1;
				end if;
				if cell4 = '1' then
					temp2 := temp2 + 1;
				end if;
				if cell5 = '1' then
					temp2 := temp2 + 1;
				end if;
				if cell6 = '1' then
					temp2 := temp2 + 1;
				end if;
				if cell7 = '1' then
					temp2 := temp2 + 1;
				end if;
				activeCells <= temp1 + temp2;
				done_countcells <= '1';
			else
				done_countcells <= '0';
			end if;
		end if;
	end process;

	--this process applies the rules to the given adjacent cells (cell0-cell7) and currentcell
	process(applyrules, clk)
	begin
		if clk'event and clk = '1' then
			if applyrules = '1' then
				if activeCells = 3 and currentcell = '0' then
					processedcell <= '1';
				elsif (activeCells = 2 or activeCells = 3) and currentcell = '1' then
					processedcell <= '1';
				else
					processedcell <= '0';
				end if;
				done_applyrules <= '1';
			else
				done_applyrules <= '0';
			end if;
		end if;
	end process;


	--this process writes from processedcell to the new array
	process(writecell, clk)
	begin
		if clk'event and clk = '1' then
			if writecell = '1' then
				gameArray(other_array)(x)(y) <= processedcell;
				done_writecell <= '1';
			elsif initialize = '1' then
				gameArray(0) <= (x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000");
				gameArray(1) <= (x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000", x"00000000000000000000000000000000");

				-- glider gun
				gameArray(0)(62)(26) <= '1';
				gameArray(0)(61)(26 downto 23) <= "1111";
				gameArray(0)(61)(31) <= '1';
				gameArray(0)(60)(14) <= '1';
				gameArray(0)(60)(25 downto 22) <= "1111";
				gameArray(0)(60)(31) <= '1';
				gameArray(0)(59)(13) <= '1';
				gameArray(0)(59)(15) <= '1';
				gameArray(0)(59)(22) <= '1';
				gameArray(0)(59)(25) <= '1';
				gameArray(0)(59)(36 downto 35) <= "11";
				gameArray(0)(58)(12) <= '1';
				gameArray(0)(58)(17 downto 16) <= "11";
				gameArray(0)(58)(25 downto 22) <= "1111";
				gameArray(0)(58)(36 downto 35) <= "11";
				gameArray(0)(57)(2 downto 1) <= "11";
				gameArray(0)(57)(12) <= '1';
				gameArray(0)(57)(17 downto 16) <= "11";
				gameArray(0)(57)(26 downto 23) <= "1111";
				gameArray(0)(56)(2 downto 1) <= "11";
				gameArray(0)(56)(12) <= '1';
				gameArray(0)(56)(17 downto 16) <= "11";
				gameArray(0)(56)(26) <= '1';
				gameArray(0)(55)(13) <= '1';
				gameArray(0)(55)(15) <= '1';
				gameArray(0)(54)(14) <= '1';

				-- r pentimino
				gameArray(0)(43)(63) <= '1';
				gameArray(0)(44)(63) <= '1';
				gameArray(0)(44)(62) <= '1';
				gameArray(0)(45)(62) <= '1';
				gameArray(0)(44)(64) <= '1';

				-- puff gliders
				gameArray(0)(20)(99 downto 96) <= "1111";
				gameArray(0)(21)(100) <= '1';
				gameArray(0)(21)(96) <= '1';
				gameArray(0)(22)(96) <= '1';
				gameArray(0)(23)(97) <= '1';
				gameArray(0)(26)(99) <= '1';
				gameArray(0)(27)(98) <= '1';
				gameArray(0)(28)(98) <= '1';
				gameArray(0)(29)(99 downto 98) <= "11";
				gameArray(0)(30)(100) <= '1';
				gameArray(0)(34)(99 downto 96) <= "1111";
				gameArray(0)(35)(100) <= '1';
				gameArray(0)(35)(96) <= '1';
				gameArray(0)(36)(96) <= '1';
				gameArray(0)(37)(97) <= '1';

				done_initialize <= '1';
			elsif copy_over = '1' then
				gameArray(0) <= gameArray(1);
				done_copy <= '1';
				done_writecell <= '0';
			else
				done_writecell <= '0';
				done_initialize <= '0';
				done_copy <= '0';
			end if;
		end if;
	end process;
end life_arch;
